#!/usr/bin/env ruby

#
# WPScan - WordPress Security Scanner
# Copyright (C) 2011  Ryan Dewhurst AKA ethicalhack3r
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ryandewhurst at gmail
#

# This library should contain all methods used to gather (discover) information
# from a wordpress installation.

class Discover

  attr_accessor :url, :threads, :verbose

  def initialize(url, threads, verbose)
    @url = url
    @threads = threads
    @verbose = verbose
  end

  # Checks to see if the readme.html file exists
  #
  # This file comes by default in a wordpress installation,
  # and if deleted is reinstated with an upgrade.

  def readme_exists?

    exists = false

    response = Typhoeus::Request.get(@url.to_s + 'readme.html')
    
    if response.body =~ %r{Semantic Personal Publishing Platform}i
      exists = true
    end

    exists

  end

  # Attempts to find the wordpress version from,
  # the generator meta tag in the html source.
  #
  # The meta tag can be removed however it seems, 
  # that it is reinstated on upgrade.

  def generator_version

    version = nil

    response = Typhoeus::Request.get(@url.to_s, :follow_location => true, :max_redirects => 2)
    response2 = Typhoeus::Request.get(@url.to_s + "feed/", :follow_location => true, :max_redirects => 2)

    version = response.body[%r{name="generator" content="wordpress (.*)"}i, 1]

    # if the first request was unsuccessfultry the rss feed
    if version.nil?
      version = response2.body[%r{<generator>http://wordpress.org/\?v=(.*)</generator>}i, 1]
    end

    version

  end

  # Uses data/wp_versions.xml to try to identify a
  # wordpress version.
  #
  # It does this by using client side file hashing
  # with a scoring system. 
  # 
  # The scoring system is a number representing
  # the uniqueness of a client side file across
  # all versions of wordpress.
  # 
  # Example:
  #
  # Score - Hash - File - Versions
  #   1 - 3e63c08553696a1dedb24b22ef6783c3 - /wp-content/themes/twentyeleven/style.css - 3.2.1
  #   2 - 15fc925fd39bb496871e842b2a754c76 - /wp-includes/js/wp-lists.js - 2.6,2.5.1
  #   3 - 3f03bce84d1d2a169b4bf4d8a0126e38 - /wp-includes/js/autosave.js - 2.9.2,2.9.1,2.9

  def advanced_version_detection
  
    version = false
   
    data = XmlSimple.xml_in('data/wp_versions.xml', { 'KeyAttr' => 'hash' })

    data['hash'].each do |hash|

      md5 = hash['md5']
      file = hash['file'][0]
      versions = hash['versions'][0]

      response = Typhoeus::Request.get(@url.to_s+file)

      if Digest::MD5.hexdigest(response.body) == md5
        version = versions
      end

    end
    
    version

  end

  # Enumerate wordpress usernames by using Veronica Valeros's technique:
  # http://seclists.org/fulldisclosure/2011/May/493 

  def usernames

    usernames = []
   
    # Iterate through 10 wordpress user ids

    (1..10).each do |author_count|

      # build the new URL and send the request

      response = Typhoeus::Request.get(@url.to_s + '?author=' + author_count.to_s)

      if response.code == 301 # username in location? 
        usernames.push(response.headers_hash['location'][%r{/author/(.*)/}i, 1])
      elsif response.code == 200 # username in body?
        usernames.push(response.body[%r{posts by (.*) feed}i, 1])
      end

    end

    # clean the array, remove nils and possible duplicates

    usernames.flatten!
    usernames.compact!
    usernames.uniq!

    return usernames

  end

  # Compare the wordpress version to a list of vulnerabilities,
  # the vulnerability file is stored in data/wp_vuln.xml.

  def wp_vulnerabilities(version)
    data = XmlSimple.xml_in('data/wp_vulns.xml', { 'KeyAttr' => 'version' })
    vulnerabilities = data['wordpress'][version] 
  end

  # See if there are any vulnerabilities associated with the,
  # plugins which we found to be installed.    

  def plugin_vulnerabilities(plugins)

    vulnerabilities = []
    data = XmlSimple.xml_in('data/plugin_vulns.xml', { 'KeyAttr' => 'plugin' })

    data['plugin'].each do |plugin|
      if plugins.include?(plugin['name'])
        vulnerabilities << plugin
      end
    end

    vulnerabilities

  end

  # Discover the wordpress theme name by parsing the csss link rel

  def theme_name

    response = Typhoeus::Request.get(@url.to_s)

    # follow redirects...
    until response.code !~ /30/
      response = Net::HTTP.get_response(URI.parse(response.headers_hash['location']))
    end

    theme = response.body[%r{/wp-content/themes/(.*)/style.css}i, 1]
  end

  # Discover any error_log files created by WordPress
  # These are created by the WordPress error_log() function
  # They are normally found in the /plugins/ directory,
  # however can also be found in their specific plugin dir.
  # http://www.exploit-db.com/ghdb/3714/

  def error_log(plugins)

    hydra = Typhoeus::Hydra.new(:max_concurrency => @threads, :timeout => 2000)

    error_logs = []

    plugins.each do |plugin|

      url = @url.to_s + 'wp-content/plugins/' + plugin  + '/error_log'
      request = Typhoeus::Request.new(url)

      request.on_complete do |response|
        if response.body =~ %r{PHP Fatal error}i
          error_logs << url
        end
      end

      hydra.queue(request)

    end

    hydra.run

    error_logs

  end

  # Enumerate installed plugins.

  def plugins

    hydra = Typhoeus::Hydra.new(:max_concurrency => @threads, :timeout => 2000)

    found_plugins = []
    request_count = 0
    queue_count = 0
    local_404_hash = error_404_hash

    # Open and parse the 'most popular' plugin list...
    file_contents = File.open("data/plugins.txt","r") {|file| file.readlines.collect{|line| line.chomp}}

    # Open and parse the plugins we know have vulnerabilities...
    xml_contents = XmlSimple.xml_in('data/plugin_vulns.xml', { 'KeyAttr' => 'name' })

    # Join both plugin arrays and make them into a set
    plugins = file_contents | xml_contents['plugin'].keys

    plugins.each do |plugin|

      if plugin =~ /\// # has a file name
        request = Typhoeus::Request.new(@url.to_s + "wp-content/plugins/" + plugin)
      else
        request = Typhoeus::Request.new(@url.to_s + "wp-content/plugins/" + plugin + "/")
      end

     request_count += 1      

      request.on_complete do |response|

        print "\rChecking for " + plugins.size.to_s + " total plugins... #{(request_count * 100) / plugins.size}% complete." # progress indicator

        if response.code == 200 or response.code == 403
          if Digest::MD5.hexdigest(response.body) != local_404_hash
            if plugin =~ /\// # has a file name
              found_plugins.push(plugin.split('/')[0])
            else
              found_plugins.push(plugin)
            end
          end
        end

      end

      hydra.queue(request)
      queue_count += 1

      if queue_count == @threads
        hydra.run
        queue_count = 0
      end

    end

    hydra.run

    found_plugins

  end

  # Tries to identify timthumb.php files.
  # Data is in data/timthumb.txt
  # This data was collected from my own apache logs.

  def timthumb(theme_name)
  
    hydra = Typhoeus::Hydra.new(:max_concurrency => @threads, :timeout => 2000)

    found_timthumb = []
    request_count = 0
    queue_count = 0

    # Open and parse the timthumb.txt file...
    file_contents = File.open("data/timthumb.txt","r") {|file| file.readlines.collect{|line| line.chomp}}

    # Add the theme_name to the list of checks...
    unless theme_name == false or theme_name == '' or theme_name.nil?
      theme_name = URI.escape(theme_name)
      file_contents << 'wp-content/themes/' + theme_name + '/timthumb.php' 
      file_contents << 'wp-content/themes/' + theme_name + '/lib/timthumb.php'
      file_contents << 'wp-content/themes/' + theme_name + '/inc/timthumb.php'
      file_contents << 'wp-content/themes/' + theme_name + '/includes/timthumb.php'
      file_contents << 'wp-content/themes/' + theme_name + '/scripts/timthumb.php'
      file_contents << 'wp-content/themes/' + theme_name + '/tools/timthumb.php'
      file_contents << 'wp-content/themes/' + theme_name + '/functions/timthumb.php'
    end

    file_contents.each do |line|

      request = Typhoeus::Request.new(@url.to_s + line)
     
      request_count += 1

      request.on_complete do |response|
        print "\rChecking for " + file_contents.size.to_s + " total timthumb files... #{(request_count * 100) / file_contents.size}% complete." # progress indicator
        if response.body =~ /no image specified/i
          found_timthumb.push(line)
        end
      end

      hydra.queue(request)
      queue_count += 1

      if queue_count == @threads
        hydra.run
        queue_count = 0
      end

    end

    hydra.run

    found_timthumb
 
  end

  # Is directory listing enabled?
  # WordPress denies directory listing however,
  # forgets about the plugin directory.

  def directory_listing?(plugin)

    listing = false

    response = Typhoeus::Request.get(@url.to_s + 'wp-content/plugins/' + plugin + '/')

    if response.body[%r{<title>Index of}]
      listing = true
    end

    listing

  end

  # Checks if the 'login lockdown' plugin is enabled
  # Thanks to Alip Aswalid for providing this method.

  def login_lockdown?
  
    exists = false
    
    response = Typhoeus::Request.get(@url.to_s + 'wp-login.php')

    if response.body =~ %r{Login LockDown}i
      exists = true
    end

    exists

  end

  # Check for Full Path Disclosure (FPD)
  
  def full_path_disclosure?

    fpd = false

    response = Typhoeus::Request.get(@url.to_s + 'wp-includes/rss-functions.php')

    if response.body[%r{Fatal error}i]
      fpd = true
    end

    fpd

  end

  # Return the MD5 hash of a 404 page

  def error_404_hash
    
    non_existant_page = Digest::MD5.hexdigest(rand(9999999999).to_s) + ".html"

    response = Typhoeus::Request.get(@url.to_s + non_existant_page)

    md5 = Digest::MD5.hexdigest(response.body)

  end

end
